// Approximation an MRF by projection on F0 (L2 norm) for a fixed interpretation
// Not translated since not part of the paper. 
// You'd better not use the last part of this code (not finalized).

#ifndef APPROXIMATION_F0__H
#define APPROXIMATION_F0__H

// MRF - graph-cut by Kolmogorov, Veksler et al
#include "mrf.h"
#include "ICM.h"
#include "GCoptimization.h"
#include "MaxProdBP.h"
#include "graph.h"

#include "minimization_F0.h"

namespace MRF_F0 {

// Let's reach infinity !
double infini = 100000000;

// donne les alphas de la matrice de la famille la plus proche de M, à interprétation fixe.
DoubleImage projette_alpha_L2(const DoubleImage& M){
  int ax = int( log(M.nx)/log(2) + 0.1); // ggrrr
  assert( int( pow(2,ax) + 0.1) == M.nx);

  int ay = int( log(M.ny)/log(2) + 0.1);
  assert( int( pow(2,ay) + 0.1) == M.ny);

  DoubleImage Alpha(ax,ay);
  for (int i = 0; i<ax; i++)
    for (int j = 0; j<ay; j++){
      Alpha(i,j) = matrix_to_alpha(M, i, j);
      if (Alpha(i,j) < 0){
	Alpha(i,j) = 0;
	//	cout << "#"; cout . flush();
      }
    }

  return Alpha;
}


double norme_L2_diff(const DoubleImage& A, const DoubleImage& B){
  assert(A.nx == B.nx);
  assert(A.ny == B.ny);
  double x = 0;
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      x += square(A(i,j) - B(i,j));
  return x;
}

// en donne juste le coût
double cout_projection_alpha_L2(const DoubleImage& M){
  DoubleImage Alpha = projette_alpha_L2(M);
  DoubleImage R = alphas_to_matrix(Alpha);
  Alpha.empty();
  double x = norme_L2_diff(M, R);
  R.empty();
  return x;
}








// fonctions auxiliaires pour les méthodes d'optimisation
MRF::CostVal cout_potentiel_init(int pixel, int etiq, vector< vector<double> >* Pot = NULL){
  static vector< vector<double> >* Potentiel = NULL;

  if (Pot != NULL) {
    Potentiel = Pot;
    cout << "Fonction Coût potentiel initialisée" << endl;
    return 0;
  } else {
    //    cout << "Pot(" << pixel << ", " << etiq << ")" << endl;
    assert(Potentiel != NULL);
    if (pixel< int(Potentiel->size()))
      return (*Potentiel)[pixel][etiq];
    else {
      cout << BoldBlue << "p" << TextNormal <<  "(" << pixel << ")"; cout.flush();
      assert(false);
      return 1.;
    }
  }
}

MRF::CostVal cout_potentiel(int pixel, int etiq){
  return cout_potentiel_init(pixel,etiq, NULL);
}

MRF::CostVal cout_interaction_init(int pix1, int pix2, int etiq1, int etiq2, Energy* E = NULL){
  static Energy* Ener = NULL;

  if (E != NULL) {
    Ener = E;
    cout << "Fonction Coût interaction initialisée" << endl;
    return 0;
  } else {
    //    cout << "Inter(" << pix1 << ", " << pix2 << ", " << etiq1 <<  ", " << etiq2 << ").  pix_max : " << (*Ener).nb_var << ", label_max: " << (*Ener).Pot[0].size() << endl;
    assert(Ener != NULL);
    if ((pix1 >= (*Ener).nb_var) ||  (pix2 >= (*Ener).nb_var) || (pix1 < 0) || (pix2 < 0)){
      cout << BoldGreen << "H" << TextNormal << "(" << pix1 << "," << pix2 << " / " << (*Ener).nb_var <<")"; cout.flush();
      assert(false);
      return 1.;
    }
    int vv = find_value_1((*Ener).Neighbors[pix1], pix2);
    if (vv == -1){ // pas voisins
      cout << BoldRed << "v" << TextNormal << "("  << pix1 << "," << pix2 << ")"; cout.flush();
      assert(false);
      return 1.; 
    }
    int mat = (*Ener).Neighbors[pix1][vv].second;
    if (pix1 < pix2)
      return (*Ener).Interact[mat].first(etiq1,etiq2);
    
    return (*Ener).Interact[mat].first(etiq2,etiq1);
  }
}

MRF::CostVal cout_interaction(int pix1, int pix2, int etiq1, int etiq2){
  return cout_interaction_init(pix1, pix2, etiq1, etiq2, NULL);
}







// suppose a entre -1 et 1.
inline double demilog(double a){
  if (a == -1)
    return infini;
  if (a == 1)
    return -infini;
  return 0.5 * log( (1 - a) / (1 + a) );
}

// idem mais version K-L
// donne les alphas de la matrice de la famille la plus proche de M, à interprétation fixe.
DoubleImage projette_alpha_KL(const DoubleImage& M){
  int ax = int( log(M.nx)/log(2) + 0.1);
  assert( int( pow(2,ax) + 0.1) == M.nx);

  int ay = int( log(M.ny)/log(2) + 0.1);
  assert( int( pow(2,ay) + 0.1) == M.ny);

  DoubleImage poids(M.nx, M.ny);
  double tot = 0;
  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<M.ny; j++){
      poids(i,j) = exp(-M(i,j));
      tot += poids(i,j);
    }
  
  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<M.ny; j++)
      poids(i,j) /= tot;


  DoubleImage Alpha(ax,ay);
  for (int i = 0; i<ax; i++)
    for (int j = 0; j<ay; j++){ // boucle sur coeffs;

      for (int im = 0; im<M.nx; im++)
	for (int jm = 0; jm<M.ny; jm++)
	  Alpha(i,j) += ((ith_bit(i, im) + ith_bit(j,jm) == 1)?(1):(-1)) * poids(im,jm);

      Alpha(i,j) = demilog(Alpha(i,j)); // passer de K à Alpha
      if (Alpha(i,j) < 0)
	Alpha(i,j) = 0;
    }

  poids.empty();
  return Alpha;
}



// idem mais version K-L
// donne le coût de la projection de la matrice de la famille la plus proche de M, à interprétation fixe.
double cout_projection_alpha_KL(const DoubleImage& M){
  int ax = int( log(M.nx)/log(2) + 0.1);
  assert( int( pow(2,ax) + 0.1) == M.nx);

  int ay = int( log(M.ny)/log(2) + 0.1);
  assert( int( pow(2,ay) + 0.1) == M.ny);

  DoubleImage poids(M.nx, M.ny);
  double tot = 0;
  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<M.ny; j++){
      poids(i,j) = exp(-M(i,j));
      tot += poids(i,j);
    }
  
  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<M.ny; j++)
      poids(i,j) /= tot;

  double err = 0;
  double aux;
  for (int i = 0; i<ax; i++)
    for (int j = 0; j<ay; j++){ // boucle sur coeffs;
      aux = 0;
      for (int im = 0; im<M.nx; im++)
	for (int jm = 0; jm<M.ny; jm++)
	  aux += ((ith_bit(i, im) + ith_bit(j,jm) == 1)?(1):(-1)) * poids(im,jm);

      double alph = demilog(aux);
      if (alph < 0)
	alph = 0;
      //aux : K, alpha : alpha
      
      err += log(cosh(alph)) + alph * aux;
    }

  poids.empty();
  return err;
}


// Etant donnée une matrice d'interaction (sur les étiquettes 1,2,3...) et les interprétations de chaque variable (00,01 --> 1,2,3...),
// retourne la projection sur la famille pour cette interprétation fixée.
// M_orig o Px = M_alpha
DoubleImage approxime_alpha(const DoubleImage& M_orig, const vector<int>& Px, const vector<int>& Py, int distance){
  DoubleImage A = permute_x(M_orig, Px);
  A = permute_y(A, Py);
  DoubleImage Alpha;
  switch (distance) {
  case 0 : //L2
    Alpha = projette_alpha_L2(A); break;
  case 1 : //KL
    Alpha = projette_alpha_KL(A); break;
  default :
    cout << BoldRed << "Préciser une distance !" << TextNormal << endl;
  }
  A.empty();
  return Alpha;
}


// Etant donnée une matrice d'interaction (sur les étiquettes 1,2,3...) et les interprétations de chaque variable (00,01 --> 1,2,3...),
// retourne le coût de la projection sur la famille pour cette interprétation fixée.
// M_orig o Px = M_alpha
double cout_projection_interaction(int distance, const DoubleImage& M_orig, const vector< vector<int> >& MetaPerm, int i, int j){
  DoubleImage A = permute_x(M_orig, MetaPerm[i]);
  A = permute_y(A, MetaPerm[j]);
  double x;
  switch (distance) {
  case 0 : //L2
    x = cout_projection_alpha_L2(A); break;
  case 1 : //KL
    x = cout_projection_alpha_KL(A); break;
  default :
    cout << BoldRed << "Préciser une distance !" << TextNormal << endl;
  }
  A.empty();
  return x;
}



// Etant donné un vecteur de potentiel, retourne la projection sur la famille pour cette interprétation.
// version Kullback-Leibler
vector<double> projette_potentiel_KL(const vector<double>& V){
  int n_etiq = int(V.size());
  int n = int( log(n_etiq)/log(2) + 0.1);
  assert( int( pow(2,n) + 0.1) == n_etiq);
  int n_tot = (n * (n+1)) / 2;

  // poids expo-négatif
  vector<double> poids(n_etiq);
  double tot = 0;
  for (int i = 0; i<n_etiq; i++){
    poids[i] = exp(-V[i]);
    tot += poids[i];
  }
  for (int i = 0; i<n_etiq; i++)
    poids[i] /= tot;


  vector<double> alpha(n_tot,0);

  for (int c = 0; c < n; c++){ // les coeffs dans R
    for (int i = 0; i < n_etiq; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      alpha[c] += s * poids[i];
    }

    alpha[c] = demilog( alpha[c] );
  }
  

  int cn = n;
  for (int c1 = 0; c1 < n; c1++) // les coeffs dans R+
    for (int c2 = c1+1; c2 < n; c2++){

      for (int i = 0; i < n_etiq; i++){
        int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
        alpha[cn] += s * poids[i];
      }

      alpha[cn] = demilog( alpha[cn] );

      // le coeff doit être > 0.
      if (alpha[cn] < 0)
	alpha[cn] = 0;

      cn++;
    }
  assert(cn == n_tot);


  return alpha;
}




// Etant donné un vecteur de potentiel, retourne le coût de la projection sur la famille pour cette interprétation.
// version Kullback-Leibler
double cout_projection_potentiel_KL(const vector<double>& V){

  int n_etiq = int(V.size());
  int n = int( log(n_etiq)/log(2) + 0.1);
  assert( int( pow(2,n) + 0.1) == n_etiq);
  int n_tot = (n * (n+1)) / 2;

  // poids expo-négatif
  vector<double> poids(n_etiq);
  double tot = 0;
  for (int i = 0; i<n_etiq; i++){
    poids[i] = exp(-V[i]);
    tot += poids[i];
  }
  for (int i = 0; i<n_etiq; i++)
    poids[i] /= tot;

  double somme = 0;
  double alpha;

  for (int c = 0; c < n; c++){ // les coeffs dans R
    alpha = 0;
    for (int i = 0; i < n_etiq; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      alpha += s * poids[i];
    }
    // ici, alpha = K, et al = alpha
    double al = demilog( alpha) ;
    alpha = log(cosh(al)) + al * alpha; // et maintenant alpha c'est le cout
    somme += alpha;
  }
  

  int cn = n;
  for (int c1 = 0; c1 < n; c1++) // les coeffs dans R+
    for (int c2 = c1+1; c2 < n; c2++){
      alpha = 0;
      for (int i = 0; i < n_etiq; i++){
        int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
        alpha += s * poids[i];
      }

      double al = demilog( alpha );

      // le coeff doit être > 0.
      if (al < 0)
	al = 0;
      
      alpha = log(cosh(al)) + al * alpha; // et maintenant alpha c'est le cout
      somme += alpha;
      cn++;
    }
  assert(cn == n_tot);

  poids.clear();

  return somme;
}




// Etant donné un vecteur de potentiel, retourne la projection sur la famille pour cette interprétation.
vector<double> projette_potentiel_L2(const vector<double>& V){
  int n_etiq = int(V.size());
  int n = int( log(n_etiq)/log(2) + 0.1);
  assert( int( pow(2,n) + 0.1) == n_etiq);
  int n_tot = (n * (n+1)) / 2;

  vector<double> alpha(n_tot,0);

  for (int c = 0; c < n; c++){ // les coeffs dans R
    for (int i = 0; i < n_etiq; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      alpha[c] += s * V[i];
    }
    alpha[c] /= n_etiq;
  }


  int cn = n;
  for (int c1 = 0; c1 < n; c1++) // les coeffs dans R+
    for (int c2 = c1+1; c2 < n; c2++){

      for (int i = 0; i < n_etiq; i++){
        int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
        alpha[cn] += s * V[i];
      }

      // le coeff doit être > 0.
      if (alpha[cn] < 0)
	alpha[cn] = 0;
      else
	alpha[cn] /= n_etiq;
      cn++;
    }
  assert(cn == n_tot);


  return alpha;
}


double norme_diff(const vector<double>& A, const vector<double>& B){
  double x = 0;
  assert(A.size() == B.size());
  for (int i = 0; i<int(A.size()); i++)
    x += square(A[i] - B[i]);
  return sqrt(x);
}

// cout de projection L2
double cout_projection_potentiel_L2(const vector<double>& V){
  vector<double> A = projette_potentiel_L2(V);
  vector<double> W = alphas_to_potential(A);
  double x = norme_diff(V,W);
  A.clear(); W.clear();
  return x;
}


// renvoie les couts de la projection du potentiel
double cout_projection_potentiel(int distance, const vector<double>& V, const vector<int>& P){
  vector<double> VP = permute(V, P);
  double alpha;
  switch (distance) {
  case 0 : // L2
    alpha = cout_projection_potentiel_L2(VP); break;
  case 1 : // K-B
    alpha = cout_projection_potentiel_KL(VP); break;
  default :
    cout << BoldRed << "Préciser une distance !" << TextNormal << endl;
  }
  VP.clear();
  return alpha;
}


double cout_projection(const Energy& E, int distance){
  double c = 0;
  for (int i = 0; i<E.nb_var; i++)
    for (int k = 0; k<int(E.Pot[i].size()); k++)
      c += cout_projection_potentiel(distance, E.Pot[i], identity_permutation( int(E.Pot[i].size()) ) ); 
  

  for (int m = 0; m<int(E.Interact.size()); m++){
    vector< vector<int> > P(2);
    P[0] = identity_permutation(  int(E.Pot[ E.Interact[m].second.first  ].size()) );
    P[1] = identity_permutation(  int(E.Pot[ E.Interact[m].second.second  ].size()) );
    c += cout_projection_interaction(distance, E.Interact[m].first, P, 0, 1);
    P.clear();
  }
  return c;
}    



// Etant donné un vecteur de potentiel et une interprétation fixée, retourne la projection sur la famille pour cette interprétation.
// V_orig o P = V_alpha
// distance : 0 = L2, 1 = Kullback-Leibler
vector<double> approxime_potentiel(const vector<double>& V, const vector<int>& P, int distance){
  vector<double> VP = permute(V, P);
  vector<double> alpha;
  switch (distance) {
  case 0 : // L2
    alpha = projette_potentiel_L2(VP); break;
  case 1 : // K-B
    alpha = projette_potentiel_KL(VP); break;
  default :
    cout << BoldRed << "Préciser une distance !" << TextNormal << endl;
  }
  VP.clear();
  return alpha;
}


/***** WARNING : THIS APPROXIMATION IS VERY SIMPLE *****/
// Builds an approximation of the energy by projecting it on the family F0 
// (with permutations = identity, i.e. for fixed interpretation)
// Supposes that E has been canonized previously.
// distance : 0 = L2, 1 = Kullback-Leibler
void construct_approximation(PreGraph& PG, const Energy& E, int distance = 0, bool verbose_basic = true){
  vector<DoubleImage>& Alpha = PG.Alpha;
  vector< vector<int> >& Interp = PG.Interp;
  vector< vector<double> >& Coeff_Pot = PG.Coeff_Pot;
  vector<bool>& solitary = PG.solitary;

  Alpha = vector<DoubleImage>(E.Interact.size());
  Interp = vector< vector<int> >(E.nb_var);
  Coeff_Pot = vector< vector<double> >(E.nb_var);
  solitary = vector<bool>(E.nb_var, false);

  for (int i = 0; i<E.nb_var; i++){
    // Gère les solitary
    if (E.Neighbors[i].size() == 0){
      solitary[i] = true;
      continue;
    }
    
    // Fixe les interprétations à l'identité
    Interp[i] = identity_permutation(E.Pot[i].size());
    
    // Calcule les Coeff_Pot
    Coeff_Pot[i] = approxime_potentiel(E.Pot[i],  Interp[i], distance);
  }
  
  for (int k = 0; k<int(E.Interact.size()); k++)
    // Calcule les Alpha // Il faut que les Interact[k] soient de la forme (M_ij, i, j) avec i<j.
    Alpha[k] = approxime_alpha(  E.Interact[k].first,   Interp[ E.Interact[k].second.first ],   Interp[ E.Interact[k].second.second ]  , distance);


  if (verbose_basic){
    int a = 0;
    for (int i = 0; i<E.nb_var; i++)
      if (solitary[i])
	a++;
    cout << "Nb of solitary variables : " << a << endl;
  }

}




/******************************************************************************************************************/
/****************************************** !!!   WARNING   !!! ***************************************************/
/******************************************************************************************************************/
/************************* THE CODE HEREAFTER IS NOT FINALIZED AND SHOULD NOT BE USED *****************************/
/******************************************************************************************************************/


/********* Test of projections above on random energies *********/
// (peforms rather poorly, more experimentations needed)
// Distance : 0 = L2, 1 = KL
void minimise_energie_pas_familiale_aleatoire(int distance, bool verbose_basic = true, bool verbose_a_lot = false, bool verbose_graph = false, bool test_stupide = true, bool test_local = false, bool test_global = false, ostream& cout = std::cout){

  if (verbose_basic){
    cout << TextGreen << "**********************************************************" << endl;
    cout << "***************" << TextRed << "  Création d'une énergie  " << TextGreen << "*****************" << endl;
    cout << "**********************************************************" << TextNormal << endl;
  }

  /********* Paramètres de création automatique d'énergie aléatoire ********/

  int nb_var = 2 + ( rand() % 500 ) ; // nb variables x_i
  if (verbose_basic)
    cout << "Nb variables : " << nb_var << endl;
  vector< int > NbEtiq(nb_var); // nb étiquettes possibles pour chaque variable x_i   // pas très utile, c'est Pot[i].size()

  for (vector<int>::iterator i = NbEtiq.begin(); i != NbEtiq.end(); ++i)
    (*i) = 4; //int( pow(2, 1 + (rand() % 3)  ) + 0.1 ); 
  
  int mean_neighborhood_size = 4; // à multiplier par 2 pour avoir le nb moyen de voisins d'une variable.
  double amplitude_potentiel = 1.; // amplitude du potentiel APRES canonisation.

  /******** Création potentiels, matrices et permutations ********/

  vector< vector<double> > Pot(nb_var); // les potentiels de chaque étiquette en chaque variable
  for (int i = 0; i<nb_var; i++){
    //      NbEtiq[i] = 8;
    //      Pot[i] = vector<double>(NbEtiq[i], 0); // initialisés à 0
    Pot[i] = random_vector(NbEtiq[i], 1); // coeff d'amplitude des valeurs
    //Pot[i] = generate_possible_potential(NbEtiq[i]);
    if (verbose_a_lot)
      cout << "Pot x_" << i << " : " << Pot[i] << endl;
    // = vector<double>(NbEtiq[i], 0);
    //for (vector<double>::iterator j = Pot[i].begin(); j != Pot[j].end(); ++j)
    //	(*j) = rand() * 1. / RAND_MAX;
  }
    
  vector< vector< pair<int, int> > > Neighbors(nb_var); // liste, pour chaque variable, des index de ses voisins (avec l'index de la matrice d'interaction associée), initialisée à "aucun voisin"
  //  typedef vector< pair<DoubleImage, pair<int,int> > >   MIV; // type pour ce qui suit
  Energy::MIV Interact; // matrices d'interaction avec les numéros des variables concrenées
  int nb_mat_inter = 0;
  int aux_rand = (nb_var-1)/(2*mean_neighborhood_size);

  //  // une permutation aléatoire par variable x_i
  //  vector< vector<int> > permix(nb_var);
  //  for (int i = 0; i<nb_var; i++){
  //    //    if (i != 0) //== nb_var/2)
  //    permix[i] = generate_permutation(NbEtiq[i]);
  //    if (verbose_a_lot)
  //      cout << "Perm x_" << i << " : " << permix[i] << endl;
  //    //    else
  //    // permix[i] = identity_permutation(NbEtiq[i]);
  //    Pot[i] = permute(Pot[i], permix[i]);
  //    if (verbose_a_lot)
  //      cout << "Pot' x_" << i << " : " << Pot[i] << endl;
  //  }

  for (int i = 0; i<nb_var; i++)
    for (int j = i+1; j<nb_var; j++)
      if ((aux_rand < 2) || (rand()%(aux_rand) == 0)){
	DoubleImage M;
	//if (rand()%200 == 0)
	M = random_matrix(NbEtiq[i], NbEtiq[j], 1); // 10 : coeff d'importance
	//else{
	//                                                                                        permut'qcq, ordonne alphas
	//M = generate_possible_matrix( int( log(NbEtiq[i]) /log(2) + 0.1), int( log(NbEtiq[j]) / log(2) + 0.1) , false  , true);
	//	if (verbose_a_lot){ cout << "M_" << i << "," << j << ":" << endl << M; } //; M.display_text(); }
	//	M = permute_x(M, permix[i]);
	//	//	cout << "Après perm_x_i : M_" << i << "," << j << ":" << endl; M.display_text();
	//	M = permute_y(M, permix[j]);
	//	//	cout << "Après perm_x_j : M_" << i << "," << j << ":" << endl; M.display_text();
	//	//	}
	
	Interact.push_back( make_pair(M, make_pair(i,j)) ); // convention : numéros des variables dans l'ordre croissant, ie i<j => M = Mij et non pas Mji
	Neighbors[i].push_back( make_pair(j, nb_mat_inter) );
	Neighbors[j].push_back( make_pair(i, nb_mat_inter) );
	nb_mat_inter++;
      }
  if (verbose_basic)
    cout << "Nb interactions : " << nb_mat_inter << endl;
  //  permix.clear();
  NbEtiq.clear();


  /************* L'énergie est prête ***********/

  Energy E(Pot, Neighbors, Interact, 0);

  //  Energy E0(Pot, Neighbors, Interact, 0); // Energy initiale
  //  Energy E; E.copy(E0); // Energy que l'on va canoniser

  std::cout << "."; std::cout.flush();
  

  /*********************************************************************************************************/
  /*************************   canonisation et représentation familiale  ***********************************/
  /*********************************************************************************************************/

  if (verbose_basic)
    cout << "Canonisation" << endl;

  canonise(E);
  if (amplitude_potentiel != 1){
    std::cout << endl << BoldBlue << " ********** Gros trucage !!! ************ " << TextNormal << endl;
    cout << "Amplitude potentiel après canonisation : " << amplitude_potentiel << endl;
    for (int i = 0; i<E.nb_var; i++)
      for (int j = 0; j<int(E.Pot[i].size()); j++)
	E.Pot[i][j] *= amplitude_potentiel;
  }

  //    cout << "Energy E0 : " << endl;
  //    E0.display_text();
  
  //    cout << endl << endl << "Energy E : " << endl;
  //    E.display_text();
  
  //    /******* teste si la canonisation a changé l'énergie ******/
  //  if (test_canonisation){
  //    if (verbose_basic)
  //      cout << "Test canonisation correcte ?..." << endl;
  //    
  //    if (! test_aleatoire_egalite_energies(E0, E))
  //      cout << endl << BoldRed << "Canonisation ratée..." << TextNormal << endl;
  //  }
  //    
  //  E0.empty();




  /****** retrouve alphas et permutations possibles ******/

  //  PreGraphMulti PG;
  //  bool verbose_multi = false; bool verbose_coherence = false; bool verbose_coherence_pot = false; bool verbose_global = false;
  //  construit_pregraphe(PG, E, verbose_basic, verbose_multi, verbose_coherence, verbose_coherence_pot, verbose_global, cout);

  PreGraph PG;
  PreGraph PG2;
  if (distance < 2)
    construct_approximation(PG, E, distance, verbose_basic);
  else {
    construct_approximation(PG, E, 0, verbose_basic);
    construct_approximation(PG2, E, 1, verbose_basic);
  }

  if (verbose_basic)
    switch (distance) {
    case 0: 
      cout << "Distance L2 pour la projection." << endl; break;
    case 1:
      cout << "Divergence de Kullback-Leibler pour la projection." << endl; break;
    case 2 :
      cout << "Distance L2 + divergence de Kullback-Leibler (2 tests séparés)." << endl; break;
    default :
      cout << "Pas de distance ? ou plusieurs ?" << endl;
    }
  
  // if (!verbose_basic)
  //   switch (PG.Solutions.size()) {
  //   case 0 : // en rouge
  //     std::cout << BoldRed << PG.Solutions.size() << ";" << TextNormal; 
  //     break;
  //   case 1:
  //     std::cout << ";";
  //     break;
  //   default:
  //     std::cout << BoldGreen << PG.Solutions.size() << ";" << TextNormal; 
  //   }
  // std::cout.flush();

  // if (PG.Solutions.empty()){
  //   if (verbose_basic)
  //     cout << "Energy pas familiale" << endl;
  //
  // } else {
  //   if (verbose_basic)
  //     cout << "Energy familiale !" << endl;
  // }

  // if (PG.Solutions.front().size() == 0){
  //   std::cout << BoldRed << "Mais heu ?" << TextNormal << endl;
  // }

  //  return;

  /*********************************************************************************************************/
  /***********************************   graphe et minimisation   ******************************************/
  /*********************************************************************************************************/


  if (verbose_basic){ cout << endl;
    cout << TextGreen << "**********************************************************" << endl;
    cout << "*********" << TextRed << "  Création du graphe et minimisation  " << TextGreen << "***********" << endl;
    cout << "**********************************************************" << TextNormal << endl;
  }

  vector<int> Resultat = minimize_familial_MRF(PG, E,  test_stupide, test_local, test_global, verbose_basic, verbose_graph, verbose_a_lot);
  vector<int> Resultat2;
  if (distance == 2)
    Resultat2 = minimize_familial_MRF(PG2, E,  test_stupide, test_local, test_global, verbose_basic, verbose_graph, verbose_a_lot);


  //  vector<int> Resultat = minimize_familial_MRF(E, PG, true, verbose_graph);
  
  // Nettoyages
  PG.empty();
  E.empty();
  Resultat.clear();
  if (distance == 2)
    Resultat2.clear();
  
  //    for (int k = 0; k<int( AlpesL.size() ); k++)
  //      for (int i = 0; i<int(AlpesL[k].size()); i++)
  //	AlpesL[k][i].empty();
  //    AlpesL.clear();
  
  //    Coherence.clear();
  //    Coeff_Pot.clear();
}




// Mettre des exp partout
void  prepare_energie_pour_LBP(Energy& E){

  for (int m = 0; m < int(E.Interact.size() ); m++){
    DoubleImage& M = E.Interact[m].first;
    for (int i = 0; i<M.nx; i++)
      for (int j = 0; j<M.ny; j++)
	M(i,j) = exp(-1. * M(i,j));
  }

  double a = E.Pot[0][0]; // max Pot
  for (int i = 0; i< E.nb_var; i++)
    for (int k = 0; k<int(E.Pot[k].size()); k++)
      if (a < E.Pot[i][k])
	a = E.Pot[i][k];

  a = - 1. / (4. * a);
  for (int i = 0; i< E.nb_var; i++)
    for (int k = 0; k<int(E.Pot[k].size()); k++)
      E.Pot[i][k] = exp(a * E.Pot[i][k] );

}




// ajoute des constantes à tous les poids d'une énergie pour qu'ils soient positifs.
void positive(Energy& E){
  for (int i = 0; i<E.nb_var; i++){
    double a = E.Pot[i][ find_smallest(E.Pot[i]) ];
    for (int j = 0; j< int( E.Pot[i].size() ); j++)
      E.Pot[i][j] -= a;
    E.constante += a;
  }

  for (int m = 0; m< int(E.Interact.size()); m++){
    double a = smallest_coeff(E.Interact[m].first);
    for (int i = 0; i<E.Interact[m].first.nx; i++)
      for (int j = 0; j<E.Interact[m].first.ny; j++)
	(E.Interact[m].first)(i,j) -= a;
    E.constante += a;
  }
}




/********** Warning : this code relies on Loopy Belief Propagation ************/
// which is not an exact optimization tool, but only approximate.
// It turns out that the particular implementation of LBP used performs very poorly on this kind of energies.
// So, basically, forget this.

// Minimizes an MRF by Loopy Belief Propagation
// nb_etiq = idem for all variables
// Be careful in the way of indexing variables (width, length)
// Give you MRF, we'll make it positive et exponentiate it.
vector<int> minimise_LBP(const Energy& E0, int largeur, int longueur, bool verbose_basic, bool verbose_a_lot, bool test_stupide, int nbfois = 30){

  Energy Exp1; Exp1.copy(E0);

  int nb_var = Exp1.nb_var;
  int nb_etiq = int( Exp1.Pot[0].size() ); // suppose toutes la variables avec le même nb d'étiquettes


  positive(Exp1); // avec tous coeff > 0 (avec des + cst)

  /* // Energy bête
     for (int i = 0; i<E0.nb_var; i++) // test stupide
    for (int k = 0; k<int(E0.Pot[i].size()); k++)
      E0.Pot[i][k] = 0.1;

   for (int m = 0; m<int(E0.Interact.size()); m++){
     DoubleImage& M = E0.Interact[m].first;
     for (int i = 0; i<M.nx; i++)
       for (int j = 0; j<M.ny; j++){
	 int a = ((E0.Interact[m].second.first)%largeur)%10;
	 int b = ((E0.Interact[m].second.second)%largeur)%10;
	 if (  ((a == 5) && (b == 6)) || ((a == 6)&&(b == 5) ) )
	   M(i,j) = sqrt(square((int)E0.Pot[i].size()) - square(i-j)); // + 0.001 * (rand()%10) ); // + (rand()% 4) );
	 else
	   M(i,j) = 0.1 * sqrt(square(i-j) + 0.001 * (rand()%10) );
       }
   }

  E0.constante = 0;
  */

    //  Energy Exp1; Exp1.copy(E0);
  prepare_energie_pour_LBP(Exp1); // on met des exp partout, attention ce n'est plus une énergie !

  // Initialisation des coûts
  cout_potentiel_init(0,0,&Exp1.Pot); // les 0 ne servent à rien.
  cout_interaction_init(0,0,0,0,&Exp1);

  vector<int> Resultat(nb_var);
  int fois = 0;
  //for (fois = 0; fois < 2; fois++) 
  {
  DataCost *data = new DataCost(cout_potentiel);
  SmoothnessCost *smooth = new SmoothnessCost(cout_interaction); 
  EnergyFunction *eng    = new EnergyFunction(data,smooth);

  float t;
  MRF* mrf = new MaxProdBP(largeur,longueur,nb_etiq,eng);

  mrf->initialize();  
  mrf->clearAnswer();

  //  int nbfois = 60;
  if (fois == 1)
    nbfois = 200;
  mrf->optimize(nbfois,t);  // run for 5 iterations, store time t it took 
  MRF::EnergyVal E_smooth = mrf->smoothnessEnergy();
  MRF::EnergyVal E_data   = mrf->dataEnergy();
  if (verbose_basic)
    cout << endl << "Total Energy = " << E_smooth+E_data << " (Smoothness energy "  << E_smooth << ", Data Energy " << E_data <<  ")" << endl;

  cout << endl;
  if (verbose_a_lot){
    for (int pix = 0; pix < nb_var; pix++ ){
      if ((pix % largeur) == 0)
	cout << endl;
      cout << mrf->getLabel(pix) << " ";
    }
  }
  //printf("Label of pixel %d is %d",pix, mrf->getLabel(pix));
  cout << endl;
  
  for (int pix = 0; pix < nb_var; pix++ )
    Resultat[pix] = mrf->getLabel(pix);
  
  double r_lbp = evaluate_energy(E0, Resultat);
  //  cout << endl << "Résultat Loopy Belief Propagation Exp1 : " << evaluate_energy(Exp1, Resultat) << endl;
  cout << endl << "Résultat Loopy Belief Propagation E0   : " << r_lbp << endl;

  delete mrf;

  }


  if (test_stupide){
    vector<int> Coord(nb_var);
    for (int i = 0; i<nb_var; i++)
      Coord[i] = find_smallest(E0.Pot[i]); // étiquette déjà réelle, pas binaire dans ce cas                                                                        
    double r_stupide = evaluate_energy(E0, Coord);
    cout << endl << "Test des variables optimales (très) localement E0 : " << r_stupide << endl;

    Coord.clear();
    // for (int i = 0; i<nb_var; i++)
    //   Coord[i] = trouve_plus_grand(Exp1.Pot[i]); // étiquette déjà réelle, pas binaire dans ce cas                                                                        
    // r_stupide = evaluate_energy(Exp1, Coord);
    // cout << endl << "Test des variables optimales (très) localement Exp1 max : " << r_stupide << endl;

  }

  Exp1.empty();

  return Resultat;
  
}

} // end namespace
#endif
